import ctypes.wintypes

from .error import handle_nonzero_success
from .api import filesystem


def DeviceIoControl(device, io_control_code, in_buffer, out_buffer, overlapped=None):
    if overlapped is not None:
        raise NotImplementedError("overlapped handles not yet supported")

    if isinstance(out_buffer, int):
        out_buffer = ctypes.create_string_buffer(out_buffer)

    in_buffer_size = len(in_buffer) if in_buffer is not None else 0
    out_buffer_size = len(out_buffer)
    assert isinstance(out_buffer, ctypes.Array)

    returned_bytes = ctypes.wintypes.DWORD()

    res = filesystem.DeviceIoControl(
        device,
        io_control_code,
        in_buffer,
        in_buffer_size,
        out_buffer,
        out_buffer_size,
        returned_bytes,
        overlapped,
    )

    handle_nonzero_success(res)
    handle_nonzero_success(returned_bytes)

    return out_buffer[: returned_bytes.value]



import os.path


# realpath taken from https://bugs.python.org/file38057/issue9949-v4.patch
def realpath(path):
    if isinstance(path, str):
        prefix = '\\\\?\\'
        unc_prefix = prefix + 'UNC'
        new_unc_prefix = '\\'
        cwd = os.getcwd()
    else:
        prefix = b'\\\\?\\'
        unc_prefix = prefix + b'UNC'
        new_unc_prefix = b'\\'
        cwd = os.getcwdb()
    had_prefix = path.startswith(prefix)
    path, ok = _resolve_path(cwd, path, {})
    # The path returned by _getfinalpathname will always start with \\?\ -
    # strip off that prefix unless it was already provided on the original
    # path.
    if not had_prefix:
        # For UNC paths, the prefix will actually be \\?\UNC - handle that
        # case as well.
        if path.startswith(unc_prefix):
            path = new_unc_prefix + path[len(unc_prefix) :]
        elif path.startswith(prefix):
            path = path[len(prefix) :]
    return path


def _resolve_path(path, rest, seen):  # noqa: C901
    # Windows normalizes the path before resolving symlinks; be sure to
    # follow the same behavior.
    rest = os.path.normpath(rest)

    if isinstance(rest, str):
        sep = '\\'
    else:
        sep = b'\\'

    if os.path.isabs(rest):
        drive, rest = os.path.splitdrive(rest)
        path = drive + sep
        rest = rest[1:]

    while rest:
        name, _, rest = rest.partition(sep)
        new_path = os.path.join(path, name) if path else name
        if os.path.exists(new_path):
            if not rest:
                # The whole path exists.  Resolve it using the OS.
                path = os.path._getfinalpathname(new_path)
            else:
                # The OS can resolve `new_path`; keep traversing the path.
                path = new_path
        elif not os.path.lexists(new_path):
            # `new_path` does not exist on the filesystem at all.  Use the
            # OS to resolve `path`, if it exists, and then append the
            # remainder.
            if os.path.exists(path):
                path = os.path._getfinalpathname(path)
            rest = os.path.join(name, rest) if rest else name
            return os.path.join(path, rest), True
        else:
            # We have a symbolic link that the OS cannot resolve.  Try to
            # resolve it ourselves.

            # On Windows, symbolic link resolution can be partially or
            # fully disabled [1].  The end result of a disabled symlink
            # appears the same as a broken symlink (lexists() returns True
            # but exists() returns False).  And in both cases, the link can
            # still be read using readlink().  Call stat() and check the
            # resulting error code to ensure we don't circumvent the
            # Windows symbolic link restrictions.
            # [1] https://technet.microsoft.com/en-us/library/cc754077.aspx
            try:
                os.stat(new_path)
            except OSError as e:
                # WinError 1463:  The symbolic link cannot be followed
                # because its type is disabled.
                if e.winerror == 1463:
                    raise

            key = os.path.normcase(new_path)
            if key in seen:
                # This link has already been seen; try to use the
                # previously resolved value.
                path = seen[key]
                if path is None:
                    # It has not yet been resolved, which means we must
                    # have a symbolic link loop.  Return what we have
                    # resolved so far plus the remainder of the path (who
                    # cares about the Zen of Python?).
                    path = os.path.join(new_path, rest) if rest else new_path
                    return path, False
            else:
                # Mark this link as in the process of being resolved.
                seen[key] = None
                # Try to resolve it.
                path, ok = _resolve_path(path, os.readlink(new_path), seen)
                if ok:
                    # Resolution succeded; store the resolved value.
                    seen[key] = path
                else:
                    # Resolution failed; punt.
                    return (os.path.join(path, rest) if rest else path), False
    return path, True



import ctypes
from ctypes import wintypes

from .api import security
from .api import privilege
from .api import process


def get_process_token():
    """
    Get the current process token
    """
    token = wintypes.HANDLE()
    res = process.OpenProcessToken(
        process.GetCurrentProcess(), process.TOKEN_ALL_ACCESS, token
    )
    if not res > 0:
        raise RuntimeError("Couldn't get process token")
    return token


def get_symlink_luid():
    """
    Get the LUID for the SeCreateSymbolicLinkPrivilege
    """
    symlink_luid = privilege.LUID()
    res = privilege.LookupPrivilegeValue(
        None, "SeCreateSymbolicLinkPrivilege", symlink_luid
    )
    if not res > 0:
        raise RuntimeError("Couldn't lookup privilege value")
    return symlink_luid


def get_privilege_information():
    """
    Get all privileges associated with the current process.
    """
    # first call with zero length to determine what size buffer we need

    return_length = wintypes.DWORD()
    params = [
        get_process_token(),
        privilege.TOKEN_INFORMATION_CLASS.TokenPrivileges,
        None,
        0,
        return_length,
    ]

    res = privilege.GetTokenInformation(*params)

    # assume we now have the necessary length in return_length

    buffer = ctypes.create_string_buffer(return_length.value)
    params[2] = buffer
    params[3] = return_length.value

    res = privilege.GetTokenInformation(*params)
    assert res > 0, "Error in second GetTokenInformation (%d)" % res

    privileges = ctypes.cast(
        buffer, ctypes.POINTER(privilege.TOKEN_PRIVILEGES)
    ).contents
    return privileges


def report_privilege_information():
    """
    Report all privilege information assigned to the current process.
    """
    privileges = get_privilege_information()
    print("found {0} privileges".format(privileges.count))
    tuple(map(print, privileges))


def enable_symlink_privilege():
    """
    Try to assign the symlink privilege to the current process token.
    Return True if the assignment is successful.
    """
    # create a space in memory for a TOKEN_PRIVILEGES structure
    #  with one element
    size = ctypes.sizeof(privilege.TOKEN_PRIVILEGES)
    size += ctypes.sizeof(privilege.LUID_AND_ATTRIBUTES)
    buffer = ctypes.create_string_buffer(size)
    tp = ctypes.cast(buffer, ctypes.POINTER(privilege.TOKEN_PRIVILEGES)).contents
    tp.count = 1
    tp.get_array()[0].enable()
    tp.get_array()[0].LUID = get_symlink_luid()
    token = get_process_token()
    res = privilege.AdjustTokenPrivileges(token, False, tp, 0, None, None)
    if res == 0:
        raise RuntimeError("Error in AdjustTokenPrivileges")

    ERROR_NOT_ALL_ASSIGNED = 1300
    return ctypes.windll.kernel32.GetLastError() != ERROR_NOT_ALL_ASSIGNED


class PolicyHandle(wintypes.HANDLE):
    pass


class LSA_UNICODE_STRING(ctypes.Structure):
    _fields_ = [
        ('length', ctypes.c_ushort),
        ('max_length', ctypes.c_ushort),
        ('buffer', ctypes.wintypes.LPWSTR),
    ]


def OpenPolicy(system_name, object_attributes, access_mask):
    policy = PolicyHandle()
    raise NotImplementedError(
        "Need to construct structures for parameters "
        "(see http://msdn.microsoft.com/en-us/library/windows"
        "/desktop/aa378299%28v=vs.85%29.aspx)"
    )
    res = ctypes.windll.advapi32.LsaOpenPolicy(
        system_name, object_attributes, access_mask, ctypes.byref(policy)
    )
    assert res == 0, "Error status {res}".format(**vars())
    return policy


def grant_symlink_privilege(who, machine=''):
    """
    Grant the 'create symlink' privilege to who.

    Based on http://support.microsoft.com/kb/132958
    """
    flags = security.POLICY_CREATE_ACCOUNT | security.POLICY_LOOKUP_NAMES
    policy = OpenPolicy(machine, flags)
    return policy


def main():
    assigned = enable_symlink_privilege()
    msg = ['failure', 'success'][assigned]

    print("Symlink privilege assignment completed with {0}".format(msg))


if __name__ == '__main__':
    main()



import traceback
import sys
import logging

from jaraco.windows.filesystem import change

logging.basicConfig(level=logging.INFO)


def long_handler(file):
    try:
        with open(file, 'rb') as f:
            data = f.read()
        print("read", len(data), "bytes from", file)
    except Exception:
        traceback.print_exc()


def main():
    try:
        watch()
    except KeyboardInterrupt:
        pass


def watch():
    notifier = change.BlockingNotifier(sys.argv[1])
    notifier.watch_subtree = True

    for ch in notifier.get_changed_files():
        long_handler(ch)


if __name__ == '__main__':
    main()



import itertools
import contextlib

from more_itertools.recipes import consume, unique_justseen

try:
    import wmi as wmilib
except ImportError:
    pass

from jaraco.windows.error import handle_nonzero_success
from .api import power


def GetSystemPowerStatus():
    stat = power.SYSTEM_POWER_STATUS()
    handle_nonzero_success(GetSystemPowerStatus(stat))
    return stat


def _init_power_watcher():
    global power_watcher
    if 'power_watcher' not in globals():
        wmi = wmilib.WMI()
        query = 'SELECT * from Win32_PowerManagementEvent'
        power_watcher = wmi.ExecNotificationQuery(query)


def get_power_management_events():
    _init_power_watcher()
    while True:
        yield power_watcher.NextEvent()


def wait_for_power_status_change():
    EVT_POWER_STATUS_CHANGE = 10

    def not_power_status_change(evt):
        return evt.EventType != EVT_POWER_STATUS_CHANGE

    events = get_power_management_events()
    consume(itertools.takewhile(not_power_status_change, events))


def get_unique_power_states():
    """
    Just like get_power_states, but ensures values are returned only
    when the state changes.
    """
    return unique_justseen(get_power_states())


def get_power_states():
    """
    Continuously return the power state of the system when it changes.
    This function will block indefinitely if the power state never
    changes.
    """
    while True:
        state = GetSystemPowerStatus()
        yield state.ac_line_status_string
        wait_for_power_status_change()


@contextlib.contextmanager
def no_sleep():
    """
    Context that prevents the computer from going to sleep.
    """
    mode = power.ES.continuous | power.ES.system_required
    handle_nonzero_success(power.SetThreadExecutionState(mode))
    try:
        yield
    finally:
        handle_nonzero_success(power.SetThreadExecutionState(power.ES.continuous))



import ctypes.wintypes

GetModuleFileName = ctypes.windll.kernel32.GetModuleFileNameW
GetModuleFileName.argtypes = (
    ctypes.wintypes.HANDLE,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.DWORD,
)
GetModuleFileName.restype = ctypes.wintypes.DWORD



import ctypes
import os
import sys
import operator
import collections
import functools
import stat
from ctypes import (
    POINTER,
    byref,
    cast,
    create_unicode_buffer,
    create_string_buffer,
    windll,
)
from ctypes.wintypes import LPWSTR
import nt
import posixpath

from jaraco.structures import binary

from jaraco.windows.error import handle_nonzero_success
import jaraco.windows.api.filesystem as api
from jaraco.windows import reparse


def mklink():
    """
    Like cmd.exe's mklink except it will infer directory status of the
    target.
    """
    from optparse import OptionParser

    parser = OptionParser(usage="usage: %prog [options] link target")
    parser.add_option(
        '-d',
        '--directory',
        help="Target is a directory (only necessary if not present)",
        action="store_true",
    )
    options, args = parser.parse_args()
    try:
        link, target = args
    except ValueError:
        parser.error("incorrect number of arguments")
    symlink(target, link, options.directory)
    sys.stdout.write("Symbolic link created: %(link)s --> %(target)s\n" % vars())


def _is_target_a_directory(link, rel_target):
    """
    If creating a symlink from link to a target, determine if target
    is a directory (relative to dirname(link)).
    """
    target = os.path.join(os.path.dirname(link), rel_target)
    return os.path.isdir(target)


def symlink(target, link, target_is_directory=False):
    """
    An implementation of os.symlink for Windows (Vista and greater)
    """
    target_is_directory = target_is_directory or _is_target_a_directory(link, target)
    # normalize the target (MS symlinks don't respect forward slashes)
    target = os.path.normpath(target)
    flags = target_is_directory | api.SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE
    handle_nonzero_success(api.CreateSymbolicLink(link, target, flags))


def link(target, link):
    """
    Establishes a hard link between an existing file and a new file.
    """
    handle_nonzero_success(api.CreateHardLink(link, target, None))


def is_reparse_point(path):
    """
    Determine if the given path is a reparse point.
    Return False if the file does not exist or the file attributes cannot
    be determined.
    """
    res = api.GetFileAttributes(path)
    return res != api.INVALID_FILE_ATTRIBUTES and bool(
        res & api.FILE_ATTRIBUTE_REPARSE_POINT
    )


def islink(path):
    "Determine if the given path is a symlink"
    return is_reparse_point(path) and is_symlink(path)


def _patch_path(path):
    r"""
    Paths have a max length of api.MAX_PATH characters (260). If a target path
    is longer than that, it needs to be made absolute and prepended with
    \\?\ in order to work with API calls.
    See http://msdn.microsoft.com/en-us/library/aa365247%28v=vs.85%29.aspx for
    details.
    """
    if path.startswith('\\\\?\\'):
        return path
    abs_path = os.path.abspath(path)
    if not abs_path[1] == ':':
        # python doesn't include the drive letter, but \\?\ requires it
        abs_path = os.getcwd()[:2] + abs_path
    return '\\\\?\\' + abs_path


def is_symlink(path):
    """
    Assuming path is a reparse point, determine if it's a symlink.
    """
    path = _patch_path(path)
    try:
        return _is_symlink(next(find_files(path)))
    # comment below workaround for PyCQA/pyflakes#376
    except OSError as orig_error:  # noqa: F841
        orig_error.strerror = f"Error accessing {path}: {orig_error.message}"
        raise


def _is_symlink(find_data):
    return find_data.reserved[0] == api.IO_REPARSE_TAG_SYMLINK


def find_files(spec):
    r"""
    A pythonic wrapper around the FindFirstFile/FindNextFile win32 api.

    >>> root_files = tuple(find_files(r'c:\*'))
    >>> len(root_files) > 1
    True
    >>> root_files[0].filename == root_files[1].filename
    False

    This test might fail on a non-standard installation
    >>> 'Windows' in (fd.filename for fd in root_files)
    True
    """
    fd = api.WIN32_FIND_DATA()
    handle = api.FindFirstFile(spec, byref(fd))
    while True:
        if handle == api.INVALID_HANDLE_VALUE:
            raise ctypes.WinError()
        yield fd
        fd = api.WIN32_FIND_DATA()
        res = api.FindNextFile(handle, byref(fd))
        if res == 0:  # error
            error = ctypes.WinError()
            if error.winerror == api.ERROR_NO_MORE_FILES:
                break
            else:
                raise error
    # todo: how to close handle when generator is destroyed?
    # hint: catch GeneratorExit
    windll.kernel32.FindClose(handle)


def get_final_path(path):
    r"""
    For a given path, determine the ultimate location of that path.
    Useful for resolving symlink targets.
    This functions wraps the GetFinalPathNameByHandle from the Windows
    SDK.

    Note, this function fails if a handle cannot be obtained (such as
    for C:\Pagefile.sys on a stock windows system). Consider using
    trace_symlink_target instead.
    """
    desired_access = api.NULL
    share_mode = api.FILE_SHARE_READ | api.FILE_SHARE_WRITE | api.FILE_SHARE_DELETE
    security_attributes = api.LPSECURITY_ATTRIBUTES()  # NULL pointer
    hFile = api.CreateFile(
        path,
        desired_access,
        share_mode,
        security_attributes,
        api.OPEN_EXISTING,
        api.FILE_FLAG_BACKUP_SEMANTICS,
        api.NULL,
    )

    if hFile == api.INVALID_HANDLE_VALUE:
        raise ctypes.WinError()

    buf_size = api.GetFinalPathNameByHandle(hFile, LPWSTR(), 0, api.VOLUME_NAME_DOS)
    handle_nonzero_success(buf_size)
    buf = create_unicode_buffer(buf_size)
    result_length = api.GetFinalPathNameByHandle(
        hFile, buf, len(buf), api.VOLUME_NAME_DOS
    )

    assert result_length < len(buf)
    handle_nonzero_success(result_length)
    handle_nonzero_success(api.CloseHandle(hFile))

    return buf[:result_length]


def compat_stat(path):
    """
    Generate stat as found on Python 3.2 and later.
    """
    stat = os.stat(path)
    info = get_file_info(path)
    # rewrite st_ino, st_dev, and st_nlink based on file info
    return nt.stat_result(
        (stat.st_mode,)
        + (info.file_index, info.volume_serial_number, info.number_of_links)
        + stat[4:]
    )


def samefile(f1, f2):
    """
    Backport of samefile from Python 3.2 with support for Windows.
    """
    return posixpath.samestat(compat_stat(f1), compat_stat(f2))


def get_file_info(path):
    # open the file the same way CPython does in posixmodule.c
    desired_access = api.FILE_READ_ATTRIBUTES
    share_mode = 0
    security_attributes = None
    creation_disposition = api.OPEN_EXISTING
    flags_and_attributes = (
        api.FILE_ATTRIBUTE_NORMAL
        | api.FILE_FLAG_BACKUP_SEMANTICS
        | api.FILE_FLAG_OPEN_REPARSE_POINT
    )
    template_file = None

    handle = api.CreateFile(
        path,
        desired_access,
        share_mode,
        security_attributes,
        creation_disposition,
        flags_and_attributes,
        template_file,
    )

    if handle == api.INVALID_HANDLE_VALUE:
        raise ctypes.WinError()

    info = api.BY_HANDLE_FILE_INFORMATION()
    res = api.GetFileInformationByHandle(handle, info)
    handle_nonzero_success(res)
    handle_nonzero_success(api.CloseHandle(handle))

    return info


def GetBinaryType(filepath):
    res = api.DWORD()
    handle_nonzero_success(api._GetBinaryType(filepath, res))
    return res


def _make_null_terminated_list(obs):
    obs = _makelist(obs)
    if obs is None:
        return
    return u'\x00'.join(obs) + u'\x00\x00'


def _makelist(ob):
    if ob is None:
        return
    if not isinstance(ob, (list, tuple, set)):
        return [ob]
    return ob


def SHFileOperation(operation, from_, to=None, flags=[]):
    flags = functools.reduce(operator.or_, flags, 0)
    from_ = _make_null_terminated_list(from_)
    to = _make_null_terminated_list(to)
    params = api.SHFILEOPSTRUCT(0, operation, from_, to, flags)
    res = api._SHFileOperation(params)
    if res != 0:
        raise RuntimeError("SHFileOperation returned %d" % res)


def join(*paths):
    r"""
    Wrapper around os.path.join that works with Windows drive letters.

    >>> join('d:\\foo', '\\bar')
    'd:\\bar'
    """
    paths_with_drives = map(os.path.splitdrive, paths)
    drives, paths = zip(*paths_with_drives)
    # the drive we care about is the last one in the list
    drive = next(filter(None, reversed(drives)), '')
    return os.path.join(drive, os.path.join(*paths))


def resolve_path(target, start=os.path.curdir):
    r"""
    Find a path from start to target where target is relative to start.

    >>> tmp = str(getfixture('tmpdir_as_cwd'))

    >>> findpath('d:\\')
    'd:\\'

    >>> findpath('d:\\', tmp)
    'd:\\'

    >>> findpath('\\bar', 'd:\\')
    'd:\\bar'

    >>> findpath('\\bar', 'd:\\foo') # fails with '\\bar'
    'd:\\bar'

    >>> findpath('bar', 'd:\\foo')
    'd:\\foo\\bar'

    >>> findpath('\\baz', 'd:\\foo\\bar') # fails with '\\baz'
    'd:\\baz'

    >>> os.path.abspath(findpath('\\bar')).lower()
    'c:\\bar'

    >>> os.path.abspath(findpath('bar'))
    '...\\bar'

    >>> findpath('..', 'd:\\foo\\bar')
    'd:\\foo'

    The parent of the root directory is the root directory.
    >>> findpath('..', 'd:\\')
    'd:\\'
    """
    return os.path.normpath(join(start, target))


findpath = resolve_path


def trace_symlink_target(link):
    """
    Given a file that is known to be a symlink, trace it to its ultimate
    target.

    Raises TargetNotPresent when the target cannot be determined.
    Raises ValueError when the specified link is not a symlink.
    """

    if not is_symlink(link):
        raise ValueError("link must point to a symlink on the system")
    while is_symlink(link):
        orig = os.path.dirname(link)
        link = readlink(link)
        link = resolve_path(link, orig)
    return link


def readlink(link):
    """
    Return a string representing the path to which the symbolic link points.
    """
    handle = api.CreateFile(
        link,
        0,
        0,
        None,
        api.OPEN_EXISTING,
        api.FILE_FLAG_OPEN_REPARSE_POINT | api.FILE_FLAG_BACKUP_SEMANTICS,
        None,
    )

    if handle == api.INVALID_HANDLE_VALUE:
        raise ctypes.WinError()

    res = reparse.DeviceIoControl(handle, api.FSCTL_GET_REPARSE_POINT, None, 10240)

    bytes = create_string_buffer(res)
    p_rdb = cast(bytes, POINTER(api.REPARSE_DATA_BUFFER))
    rdb = p_rdb.contents
    if not rdb.tag == api.IO_REPARSE_TAG_SYMLINK:
        raise RuntimeError("Expected IO_REPARSE_TAG_SYMLINK, but got %d" % rdb.tag)

    handle_nonzero_success(api.CloseHandle(handle))
    return rdb.get_substitute_name()


def patch_os_module():
    """
    jaraco.windows provides the os.symlink and os.readlink functions.
    Monkey-patch the os module to include them if not present.
    """
    if not hasattr(os, 'symlink'):
        os.symlink = symlink
        os.path.islink = islink
    if not hasattr(os, 'readlink'):
        os.readlink = readlink


def find_symlinks(root):
    for dirpath, dirnames, filenames in os.walk(root):
        for name in dirnames + filenames:
            pathname = os.path.join(dirpath, name)
            if is_symlink(pathname):
                yield pathname
                # don't traverse symlinks
                if name in dirnames:
                    dirnames.remove(name)


def find_symlinks_cmd():
    """
    %prog [start-path]
    Search the specified path (defaults to the current directory) for symlinks,
    printing the source and target on each line.
    """
    from optparse import OptionParser
    from textwrap import dedent

    parser = OptionParser(usage=dedent(find_symlinks_cmd.__doc__).strip())
    options, args = parser.parse_args()
    if not args:
        args = ['.']
    root = args.pop()
    if args:
        parser.error("unexpected argument(s)")
    try:
        for symlink in find_symlinks(root):
            target = readlink(symlink)
            dir = ['', 'D'][os.path.isdir(symlink)]
            msg = '{dir:2}{symlink} --> {target}'.format(**locals())
            print(msg)
    except KeyboardInterrupt:
        pass


class FileAttributes(int, metaclass=binary.BitMask):
    # extract the values from the stat module on Python 3.5
    # and later.
    locals().update(
        (name.split('FILE_ATTRIBUTES_')[1].lower(), value)
        for name, value in vars(stat).items()
        if name.startswith('FILE_ATTRIBUTES_')
    )

    # For Python 3.4 and earlier, define the constants here
    archive = 0x20
    compressed = 0x800
    hidden = 0x2
    device = 0x40
    directory = 0x10
    encrypted = 0x4000
    normal = 0x80
    not_content_indexed = 0x2000
    offline = 0x1000
    read_only = 0x1
    reparse_point = 0x400
    sparse_file = 0x200
    system = 0x4
    temporary = 0x100
    virtual = 0x10000

    @classmethod
    def get(cls, filepath):
        attrs = api.GetFileAttributes(filepath)
        if attrs == api.INVALID_FILE_ATTRIBUTES:
            raise ctypes.WinError()
        return cls(attrs)


GetFileAttributes = FileAttributes.get


def SetFileAttributes(filepath, *attrs):
    """
    Set file attributes. e.g.:

        SetFileAttributes('C:\\foo', 'hidden')

    Each attr must be either a numeric value, a constant defined in
    jaraco.windows.filesystem.api, or one of the nice names
    defined in this function.
    """
    nice_names = collections.defaultdict(
        lambda key: key,
        hidden='FILE_ATTRIBUTE_HIDDEN',
        read_only='FILE_ATTRIBUTE_READONLY',
    )
    flags = (getattr(api, nice_names[attr], attr) for attr in attrs)
    flags = functools.reduce(operator.or_, flags)
    handle_nonzero_success(api.SetFileAttributes(filepath, flags))



import jaraco.windows.mmap as mmap


def test_mmap_simple():
    map = mmap.MemoryMap('foo', 50)
    with map:
        map.write(b'abc')
        map.write(b'def')
        map.seek(0)
        assert map.read(4) == b'abcd'
        assert map.read(2) == b'ef'


def test_mmap_null_bytes():
    map = mmap.MemoryMap('foo', 50)
    with map:
        map.write(b'abcdef')
        map.seek(0)
        assert map.read(10) == b'abcdef\x00\x00\x00\x00'



import ctypes
from jaraco.windows import environ
import os

getenv = ctypes.cdll.msvcrt.getenv
getenv.restype = ctypes.c_char_p
putenv = ctypes.cdll.msvcrt._putenv


def do_putenv(*pair):
    return putenv("=".join(pair))


def print_environment_variable(key):
    for method in (os.environ.get, os.getenv, environ.GetEnvironmentVariable, getenv):
        try:
            print(repr(method(key)))
        except Exception as e:
            print(e, end=' ')
    print


def do_test():
    key = 'TEST_PYTHON_ENVIRONMENT'
    print_environment_variable(key)
    methods = (
        os.environ.__setitem__,
        os.putenv,
        environ.SetEnvironmentVariable,
        do_putenv,
    )
    for i, method in enumerate(methods):
        print('round', i)
        method(key, 'value when using method %d' % i)
        print_environment_variable(key)


if __name__ == '__main__':
    do_test()



import sys
import re
import itertools
from contextlib import contextmanager
import io
import ctypes
from ctypes import windll
import textwrap
import collections

from jaraco.windows.api import clipboard, memory
from jaraco.windows.error import handle_nonzero_success
from jaraco.windows.memory import LockedMemory

__all__ = ('GetClipboardData', 'CloseClipboard', 'SetClipboardData', 'OpenClipboard')


def OpenClipboard(owner=None):
    """
    Open the clipboard.

    owner
    [in] Handle to the window to be associated with the open clipboard.
    If this parameter is None, the open clipboard is associated with the
    current task.
    """
    handle_nonzero_success(windll.user32.OpenClipboard(owner))


def CloseClipboard():
    handle_nonzero_success(windll.user32.CloseClipboard())


data_handlers = dict()


def handles(*formats):
    def register(func):
        for format in formats:
            data_handlers[format] = func
        return func

    return register


def nts(buffer):
    """
    Null Terminated String
    Get the portion of bytestring buffer up to a null character.
    """
    result, null, rest = buffer.partition('\x00')
    return result


@handles(clipboard.CF_DIBV5, clipboard.CF_DIB)
def raw_data(handle):
    return LockedMemory(handle).data


@handles(clipboard.CF_TEXT)
def text_string(handle):
    return nts(raw_data(handle))


@handles(clipboard.CF_UNICODETEXT)
def unicode_string(handle):
    return nts(raw_data(handle).decode('utf-16'))


@handles(clipboard.CF_BITMAP)
def as_bitmap(handle):
    # handle is HBITMAP
    raise NotImplementedError("Can't convert to DIB")
    # todo: use GetDIBits http://msdn.microsoft.com
    # /en-us/library/dd144879%28v=VS.85%29.aspx


@handles(clipboard.CF_HTML)
class HTMLSnippet(object):
    """
    HTML Snippet representing the Microsoft `HTML snippet format
    <https://docs.microsoft.com/en-us/windows/win32/dataxchg/html-clipboard-format>`_.
    """

    def __init__(self, handle):
        self.data = nts(raw_data(handle).decode('utf-8'))
        self.headers = self.parse_headers(self.data)

    @property
    def html(self):
        return self.data[self.headers['StartHTML'] :]

    @property
    def fragment(self):
        return self.data[self.headers['StartFragment'] : self.headers['EndFragment']]

    @staticmethod
    def parse_headers(data):
        d = io.StringIO(data)

        def header_line(line):
            return re.match(r'(\w+):(.*)', line)

        headers = map(header_line, d)
        # grab headers until they no longer match
        headers = itertools.takewhile(bool, headers)

        def best_type(value):
            try:
                return int(value)
            except ValueError:
                pass
            try:
                return float(value)
            except ValueError:
                pass
            return value

        pairs = ((header.group(1), best_type(header.group(2))) for header in headers)
        return dict(pairs)

    @classmethod
    def from_string(cls, source):
        """
        Construct an HTMLSnippet with all the headers, modeled after
        https://docs.microsoft.com/en-us/troubleshoot/cpp/add-html-code-clipboard
        """
        tmpl = textwrap.dedent(
            """
            Version:0.9
            StartHTML:{start_html:08d}
            EndHTML:{end_html:08d}
            StartFragment:{start_fragment:08d}
            EndFragment:{end_fragment:08d}
            <html><body>
            <!--StartFragment -->
            {source}
            <!--EndFragment -->
            </body></html>
            """
        ).strip()
        zeros = collections.defaultdict(lambda: 0, locals())
        pre_value = tmpl.format_map(zeros)
        start_html = pre_value.find('<html>')
        end_html = len(tmpl)
        assert end_html < 100000000
        start_fragment = pre_value.find(source)
        end_fragment = pre_value.rfind('\n<!--EndFragment')
        tmpl_length = len(tmpl) - len('{source}')
        snippet = cls.__new__(cls)
        snippet.data = tmpl.format_map(locals())
        snippet.headers = cls.parse_headers(snippet.data)
        return snippet


def GetClipboardData(type=clipboard.CF_UNICODETEXT):
    if type not in data_handlers:
        raise NotImplementedError("No support for data of type %d" % type)
    handle = clipboard.GetClipboardData(type)
    if handle is None:
        raise TypeError("No clipboard data of type %d" % type)
    return data_handlers[type](handle)


def EmptyClipboard():
    handle_nonzero_success(windll.user32.EmptyClipboard())


def SetClipboardData(type, content):
    """
    Modeled after http://msdn.microsoft.com
    /en-us/library/ms649016%28VS.85%29.aspx
    #_win32_Copying_Information_to_the_Clipboard
    """
    allocators = {
        clipboard.CF_TEXT: ctypes.create_string_buffer,
        clipboard.CF_UNICODETEXT: ctypes.create_unicode_buffer,
        clipboard.CF_HTML: ctypes.create_string_buffer,
    }
    if type not in allocators:
        raise NotImplementedError("Only text and HTML types are supported at this time")
    # allocate the memory for the data
    content = allocators[type](content)
    flags = memory.GMEM_MOVEABLE
    size = ctypes.sizeof(content)
    handle_to_copy = windll.kernel32.GlobalAlloc(flags, size)
    with LockedMemory(handle_to_copy) as lm:
        ctypes.memmove(lm.data_ptr, content, size)
    result = clipboard.SetClipboardData(type, handle_to_copy)
    if result is None:
        raise ctypes.WinError()


def set_text(source):
    with context():
        EmptyClipboard()
        SetClipboardData(clipboard.CF_TEXT, source)


def get_text():
    with context():
        result = GetClipboardData(clipboard.CF_TEXT)
    return result


def set_unicode_text(source):
    with context():
        EmptyClipboard()
        SetClipboardData(clipboard.CF_UNICODETEXT, source)


def get_unicode_text():
    with context():
        return GetClipboardData()


def get_html():
    """
    >>> set_html('<b>foo</b>')
    >>> get_html().html
    '<html><body>...<b>foo</b>...</body></html>'
    >>> get_html().fragment
    '<b>foo</b>'
    """
    with context():
        result = GetClipboardData(clipboard.CF_HTML)
    return result


def set_html(source):
    """
    >>> set_html('<b>foo</b>')
    """
    snippet = HTMLSnippet.from_string(source)
    with context():
        EmptyClipboard()
        SetClipboardData(clipboard.CF_HTML, snippet.data.encode('utf-8'))


def get_image():
    with context():
        return GetClipboardData(clipboard.CF_DIB)


def paste_stdout():
    sys.stdout.write(get_unicode_text())


def stdin_copy():
    set_unicode_text(sys.stdin.read())


@contextmanager
def context():
    OpenClipboard()
    try:
        yield
    finally:
        CloseClipboard()


def get_formats():
    with context():
        format_index = 0
        while True:
            format_index = clipboard.EnumClipboardFormats(format_index)
            if format_index == 0:
                break
            yield format_index



import ctypes.wintypes


class SYSTEM_POWER_STATUS(ctypes.Structure):
    _fields_ = (
        ('ac_line_status', ctypes.wintypes.BYTE),
        ('battery_flag', ctypes.wintypes.BYTE),
        ('battery_life_percent', ctypes.wintypes.BYTE),
        ('reserved', ctypes.wintypes.BYTE),
        ('battery_life_time', ctypes.wintypes.DWORD),
        ('battery_full_life_time', ctypes.wintypes.DWORD),
    )

    @property
    def ac_line_status_string(self):
        return {0: 'offline', 1: 'online', 255: 'unknown'}[self.ac_line_status]


LPSYSTEM_POWER_STATUS = ctypes.POINTER(SYSTEM_POWER_STATUS)
GetSystemPowerStatus = ctypes.windll.kernel32.GetSystemPowerStatus
GetSystemPowerStatus.argtypes = (LPSYSTEM_POWER_STATUS,)
GetSystemPowerStatus.restype = ctypes.wintypes.BOOL

SetThreadExecutionState = ctypes.windll.kernel32.SetThreadExecutionState
SetThreadExecutionState.argtypes = [ctypes.c_uint]
SetThreadExecutionState.restype = ctypes.c_uint


class ES:
    """
    Execution state constants
    """

    continuous = 0x80000000
    system_required = 1
    display_required = 2
    awaymode_required = 0x40



import ctypes
from .api import errors
from .api.user import GetUserName
from .error import handle_nonzero_success


def get_user_name():
    size = ctypes.wintypes.DWORD()
    try:
        handle_nonzero_success(GetUserName(None, size))
    except OSError as e:
        if e.winerror != errors.ERROR_INSUFFICIENT_BUFFER:
            raise
    buffer = ctypes.create_unicode_buffer(size.value)
    handle_nonzero_success(GetUserName(buffer, size))
    return buffer.value



import os
from path import Path


def install_pptp(name, param_lines):
    """ """
    # or consider using the API:
    # http://msdn.microsoft.com/en-us/library/aa446739%28v=VS.85%29.aspx
    pbk_path = (
        Path(os.environ['PROGRAMDATA'])
        / 'Microsoft'
        / 'Network'
        / 'Connections'
        / 'pbk'
        / 'rasphone.pbk'
    )
    pbk_path.dirname().makedirs_p()
    with open(pbk_path, 'a') as pbk:
        pbk.write('[{name}]\n'.format(name=name))
        pbk.writelines(param_lines)
        pbk.write('\n')



import subprocess


default_components = [
    'Microsoft.VisualStudio.Component.CoreEditor',
    'Microsoft.VisualStudio.Workload.CoreEditor',
    'Microsoft.VisualStudio.Component.Roslyn.Compiler',
    'Microsoft.Component.MSBuild',
    'Microsoft.VisualStudio.Component.TextTemplating',
    'Microsoft.VisualStudio.Component.VC.CoreIde',
    'Microsoft.VisualStudio.Component.VC.Tools.x86.x64',
    'Microsoft.VisualStudio.Component.VC.Tools.ARM64',
    'Microsoft.VisualStudio.Component.Windows10SDK.19041',
    'Microsoft.VisualStudio.Component.VC.Redist.14.Latest',
    'Microsoft.VisualStudio.ComponentGroup.NativeDesktop.Core',
    'Microsoft.VisualStudio.Workload.NativeDesktop',
]


def install(components=default_components):
    cmd = [
        'vs_buildtools',
        '--quiet',
        '--wait',
        '--norestart',
        '--nocache',
        '--installPath',
        'C:\\BuildTools',
    ]
    for component in components:
        cmd += ['--add', component]
    res = subprocess.Popen(cmd).wait()
    if res != 3010:
        raise SystemExit(res)


__name__ == '__main__' and install()



"""
timers
    In particular, contains a waitable timer.
"""

import time
import _thread

from jaraco.windows.api import event as win32event


class WaitableTimer:
    """
    t = WaitableTimer()
    t.set(None, 10) # every 10 seconds
    t.wait_for_signal() # 10 seconds elapses
    t.stop()
    t.wait_for_signal(20) # 20 seconds elapses (timeout occurred)
    """

    def __init__(self):
        self.signal_event = win32event.CreateEvent(None, 0, 0, None)
        self.stop_event = win32event.CreateEvent(None, 0, 0, None)

    def set(self, due_time, period):
        _thread.start_new_thread(self._signal_loop, (due_time, period))

    def stop(self):
        win32event.SetEvent(self.stop_event)

    def wait_for_signal(self, timeout=None):
        """
        wait for the signal; return after the signal has occurred or the
        timeout in seconds elapses.
        """
        timeout_ms = int(timeout * 1000) if timeout else win32event.INFINITE
        win32event.WaitForSingleObject(self.signal_event, timeout_ms)

    def _signal_loop(self, due_time, period):
        if not due_time and not period:
            raise ValueError("due_time or period must be non-zero")
        try:
            if not due_time:
                due_time = time.time() + period
            if due_time:
                self._wait(due_time - time.time())
            while period:
                due_time += period
                self._wait(due_time - time.time())
        except Exception:
            pass

    def _wait(self, seconds):
        milliseconds = int(seconds * 1000)
        if milliseconds > 0:
            res = win32event.WaitForSingleObject(self.stop_event, milliseconds)
            if res == win32event.WAIT_OBJECT_0:
                raise Exception
            if res == win32event.WAIT_TIMEOUT:
                pass
        win32event.SetEvent(self.signal_event)

    @staticmethod
    def get_even_due_time(period):
        now = time.time()
        return now - (now % period)



"""
FileChange
    Classes and routines for monitoring the file system for changes.

Copyright © 2004, 2011, 2013 Jason R. Coombs
"""

import os
import sys
import datetime
import re
from threading import Thread
import itertools
import logging

from more_itertools.recipes import consume
import jaraco.text

import jaraco.windows.api.filesystem as fs
from jaraco.windows.api import event

log = logging.getLogger(__name__)


class NotifierException(Exception):
    pass


class FileFilter(object):
    def set_root(self, root):
        self.root = root

    def _get_file_path(self, filename):
        try:
            filename = os.path.join(self.root, filename)
        except AttributeError:
            pass
        return filename


class ModifiedTimeFilter(FileFilter):
    """
    Returns true for each call where the modified time of the file is after
    the cutoff time.
    """

    def __init__(self, cutoff):
        self.cutoff = cutoff

    def __call__(self, file):
        filepath = self._get_file_path(file)
        last_mod = datetime.datetime.utcfromtimestamp(os.stat(filepath).st_mtime)
        log.debug('{filepath} last modified at {last_mod}.'.format(**vars()))
        return last_mod > self.cutoff


class PatternFilter(FileFilter):
    """
    Filter that returns True for files that match pattern (a regular
    expression).
    """

    def __init__(self, pattern):
        self.pattern = re.compile(pattern) if isinstance(pattern, str) else pattern

    def __call__(self, file):
        return bool(self.pattern.match(file, re.I))


class GlobFilter(PatternFilter):
    """
    Filter that returns True for files that match the pattern (a glob
    expression.
    """

    def __init__(self, expression):
        super(GlobFilter, self).__init__(self.convert_file_pattern(expression))

    @staticmethod
    def convert_file_pattern(p):
        r"""
        converts a filename specification (such as c:\*.*) to an equivelent
        regular expression
        >>> GlobFilter.convert_file_pattern('/*')
        '/.*'
        """
        subs = (('\\', '\\\\'), ('.', '\\.'), ('*', '.*'), ('?', '.'))
        return jaraco.text.multi_substitution(*subs)(p)


class AggregateFilter(FileFilter):
    """
    This file filter will aggregate the filters passed to it, and when called,
    will return the results of each filter ANDed together.
    """

    def __init__(self, *filters):
        self.filters = filters

    def set_root(self, root):
        consume(f.set_root(root) for f in self.filters)

    def __call__(self, file):
        return all(fil(file) for fil in self.filters)


class OncePerModFilter(FileFilter):
    def __init__(self):
        self.history = list()

    def __call__(self, file):
        file = os.path.join(self.root, file)
        key = file, os.stat(file).st_mtime
        result = key not in self.history
        self.history.append(key)
        if len(self.history) > 100:
            del self.history[-50:]
        return result


def files_with_path(files, path):
    return (os.path.join(path, file) for file in files)


def get_file_paths(walk_result):
    root, dirs, files = walk_result
    return files_with_path(files, root)


class Notifier(object):
    def __init__(self, root='.', filters=[]):
        # assign the root, verify it exists
        self.root = root
        if not os.path.isdir(self.root):
            raise NotifierException('Root directory "%s" does not exist' % self.root)
        self.filters = filters

        self.watch_subtree = False
        self.quit_event = event.CreateEvent(None, 0, 0, None)
        self.opm_filter = OncePerModFilter()

    def __del__(self):
        try:
            fs.FindCloseChangeNotification(self.hChange)
        except Exception:
            pass

    def _get_change_handle(self):
        # set up to monitor the directory tree specified
        self.hChange = fs.FindFirstChangeNotification(
            self.root, self.watch_subtree, fs.FILE_NOTIFY_CHANGE_LAST_WRITE
        )

        # make sure it worked; if not, bail
        INVALID_HANDLE_VALUE = fs.INVALID_HANDLE_VALUE
        if self.hChange == INVALID_HANDLE_VALUE:
            raise NotifierException('Could not set up directory change notification')

    @staticmethod
    def _filtered_walk(path, file_filter):
        """
        static method that calls os.walk, but filters out
        anything that doesn't match the filter
        """
        for root, dirs, files in os.walk(path):
            log.debug('looking in %s', root)
            log.debug('files is %s', files)
            file_filter.set_root(root)
            files = filter(file_filter, files)
            log.debug('filtered files is %s', files)
            yield (root, dirs, files)

    def quit(self):
        event.SetEvent(self.quit_event)


class BlockingNotifier(Notifier):
    @staticmethod
    def wait_results(*args):
        """calls WaitForMultipleObjects repeatedly with args"""
        return itertools.starmap(event.WaitForMultipleObjects, itertools.repeat(args))

    def get_changed_files(self):
        self._get_change_handle()
        check_time = datetime.datetime.utcnow()
        # block (sleep) until something changes in the
        #  target directory or a quit is requested.
        # timeout so we can catch keyboard interrupts or other exceptions
        events = (self.hChange, self.quit_event)
        for result in self.wait_results(events, False, 1000):
            if result == event.WAIT_TIMEOUT:
                continue
            index = result - event.WAIT_OBJECT_0
            if events[index] is self.quit_event:
                # quit was received; stop yielding results
                return

            # something has changed.
            log.debug('Change notification received')
            fs.FindNextChangeNotification(self.hChange)
            next_check_time = datetime.datetime.utcnow()
            log.debug('Looking for all files changed after %s', check_time)
            for file in self.find_files_after(check_time):
                yield file
            check_time = next_check_time

    def find_files_after(self, cutoff):
        mtf = ModifiedTimeFilter(cutoff)
        af = AggregateFilter(mtf, self.opm_filter, *self.filters)
        results = Notifier._filtered_walk(self.root, af)
        results = itertools.imap(get_file_paths, results)
        if self.watch_subtree:
            result = itertools.chain(*results)
        else:
            result = next(results)
        return result


class ThreadedNotifier(BlockingNotifier, Thread):
    r"""
    ThreadedNotifier provides a simple interface that calls the handler
    for each file rooted in root that passes the filters.  It runs as its own
    thread, so must be started as such::

        notifier = ThreadedNotifier('c:\\', handler = StreamHandler())
        notifier.start()
        C:\Autoexec.bat changed.
    """

    def __init__(self, root='.', filters=[], handler=lambda file: None):
        # init notifier stuff
        BlockingNotifier.__init__(self, root, filters)
        # init thread stuff
        Thread.__init__(self)
        # set it as a daemon thread so that it doesn't block waiting to close.
        # I tried setting __del__(self) to .quit(), but unfortunately, there
        # are references to this object in the win32api stuff, so __del__
        # never gets called.
        self.setDaemon(True)

        self.handle = handler

    def run(self):
        for file in self.get_changed_files():
            self.handle(file)


class StreamHandler(object):
    """
    StreamHandler: a sample handler object for use with the threaded
    notifier that will announce by writing to the supplied stream
    (stdout by default) the name of the file.
    """

    def __init__(self, output=sys.stdout):
        self.output = output

    def __call__(self, filename):
        self.output.write('%s changed.\n' % filename)



from jaraco.windows.api.filesystem import ReplaceFile

open('orig-file', 'w').write('some content')
open('replacing-file', 'w').write('new content')
ReplaceFile('orig-file', 'replacing-file', 'orig-backup', 0, 0, 0)
assert open('orig-file').read() == 'new content'
assert open('orig-backup').read() == 'some content'
import os

assert not os.path.exists('replacing-file')



import ctypes
from ctypes import WinError

from .api import memory


class LockedMemory(object):
    def __init__(self, handle):
        self.handle = handle

    def __enter__(self):
        self.data_ptr = memory.GlobalLock(self.handle)
        if not self.data_ptr:
            del self.data_ptr
            raise WinError()
        return self

    def __exit__(self, *args):
        memory.GlobalUnlock(self.handle)
        del self.data_ptr

    @property
    def data(self):
        with self:
            return ctypes.string_at(self.data_ptr, self.size)

    @property
    def size(self):
        return memory.GlobalSize(self.data_ptr)



"""cookies.py

Cookie support utilities
"""

import os
import itertools


class CookieMonster(object):
    "Read cookies out of a user's IE cookies file"

    @property
    def cookie_dir(self):
        import _winreg as winreg

        key = winreg.OpenKeyEx(
            winreg.HKEY_CURRENT_USER,
            'Software' r'\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders',
        )
        cookie_dir, type = winreg.QueryValueEx(key, 'Cookies')
        return cookie_dir

    def entries(self, filename):
        with open(os.path.join(self.cookie_dir, filename)) as cookie_file:
            while True:
                entry = itertools.takewhile(self.is_not_cookie_delimiter, cookie_file)
                entry = [item.rstrip() for item in entry]
                if not entry:
                    break
                cookie = self.make_cookie(*entry)
                yield cookie

    @staticmethod
    def is_not_cookie_delimiter(s):
        return s != '*\n'

    @staticmethod
    def make_cookie(
        key, value, domain, flags, ExpireLow, ExpireHigh, CreateLow, CreateHigh
    ):
        expires = (int(ExpireHigh) << 32) | int(ExpireLow)
        created = (int(CreateHigh) << 32) | int(CreateLow)
        flags = int(flags)
        domain, sep, path = domain.partition('/')
        path = '/' + path
        return dict(
            key=key,
            value=value,
            domain=domain,
            flags=flags,
            expires=expires,
            created=created,
            path=path,
        )



import sys
import winreg
import ctypes
import ctypes.wintypes

from jaraco.ui.editor import EditableFile

from jaraco.windows import error
from jaraco.windows.api import message, environ
from .registry import key_values as registry_key_values


def SetEnvironmentVariable(name, value):
    error.handle_nonzero_success(environ.SetEnvironmentVariable(name, value))


def ClearEnvironmentVariable(name):
    error.handle_nonzero_success(environ.SetEnvironmentVariable(name, None))


def GetEnvironmentVariable(name):
    max_size = 2**15 - 1
    buffer = ctypes.create_unicode_buffer(max_size)
    error.handle_nonzero_success(environ.GetEnvironmentVariable(name, buffer, max_size))
    return buffer.value


###


class RegisteredEnvironment(object):
    """
    Manages the environment variables as set in the Windows Registry.
    """

    @classmethod
    def show(class_):
        for name, value, type in registry_key_values(class_.key):
            sys.stdout.write('='.join((name, value)) + '\n')

    NoDefault = type('NoDefault', (object,), dict())

    @classmethod
    def get(class_, name, default=NoDefault):
        try:
            value, type = winreg.QueryValueEx(class_.key, name)
            return value
        except WindowsError:
            if default is not class_.NoDefault:
                return default
            raise ValueError("No such key", name)

    @classmethod
    def get_values_list(class_, name, sep):
        res = class_.get(name.upper(), [])
        if isinstance(res, str):
            res = res.split(sep)
        return res

    @classmethod
    def set(class_, name, value, options):
        # consider opening the key read-only except for here
        # key = winreg.OpenKey(class_.key, None, 0, winreg.KEY_WRITE)
        # and follow up by closing it.
        if not value:
            return class_.delete(name)
        do_append = options.append or (
            name.upper() in ('PATH', 'PATHEXT') and not options.replace
        )
        if do_append:
            sep = ';'
            values = class_.get_values_list(name, sep) + [value]
            value = sep.join(values)
        winreg.SetValueEx(class_.key, name, 0, winreg.REG_EXPAND_SZ, value)
        class_.notify()

    @classmethod
    def add(class_, name, value, sep=';'):
        """
        Add a value to a delimited variable, but only when the value isn't
        already present.
        """
        values = class_.get_values_list(name, sep)
        if value in values:
            return
        new_value = sep.join(values + [value])
        winreg.SetValueEx(class_.key, name, 0, winreg.REG_EXPAND_SZ, new_value)
        class_.notify()

    @classmethod
    def remove_values(class_, name, value_substring, options):
        sep = ';'
        values = class_.get_values_list(name, sep)
        new_values = [
            value for value in values if value_substring.lower() not in value.lower()
        ]
        values = sep.join(new_values)
        winreg.SetValueEx(class_.key, name, 0, winreg.REG_EXPAND_SZ, values)
        class_.notify()

    @classmethod
    def edit(class_, name, value='', options=None):
        # value, options ignored
        sep = ';'
        values = class_.get_values_list(name, sep)
        e = EditableFile('\n'.join(values))
        e.edit()
        if e.changed:
            values = sep.join(e.data.strip().split('\n'))
            winreg.SetValueEx(class_.key, name, 0, winreg.REG_EXPAND_SZ, values)
            class_.notify()

    @classmethod
    def delete(class_, name):
        winreg.DeleteValue(class_.key, name)
        class_.notify()

    @classmethod
    def notify(class_):
        """
        Notify other windows that the environment has changed (following
        http://support.microsoft.com/kb/104011).
        """
        # TODO: Implement Microsoft UIPI (User Interface Privilege Isolation) to
        #  elevate privilege to system level so the system gets this notification
        # for now, this must be run as admin to work as expected
        return_val = ctypes.wintypes.DWORD()
        res = message.SendMessageTimeout(
            message.HWND_BROADCAST,
            message.WM_SETTINGCHANGE,
            0,  # wparam must be null
            'Environment',
            message.SMTO_ABORTIFHUNG,
            5000,  # timeout in ms
            return_val,
        )
        error.handle_nonzero_success(res)


class MachineRegisteredEnvironment(RegisteredEnvironment):
    path = r'SYSTEM\CurrentControlSet\Control\Session Manager\Environment'
    hklm = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    try:
        key = winreg.OpenKey(hklm, path, 0, winreg.KEY_READ | winreg.KEY_WRITE)
    except WindowsError:
        key = winreg.OpenKey(hklm, path, 0, winreg.KEY_READ)


class UserRegisteredEnvironment(RegisteredEnvironment):
    hkcu = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)
    key = winreg.OpenKey(hkcu, 'Environment', 0, winreg.KEY_READ | winreg.KEY_WRITE)


def trim(s):
    from textwrap import dedent

    return dedent(s).strip()


def enver(*args):
    """
    %prog [<name>=[value]]

    To show all environment variables, call with no parameters:
        %prog
    To Add/Modify/Delete environment variable:
        %prog <name>=[value]

    If <name> is PATH or PATHEXT, %prog will by default append the value using
    a semicolon as a separator. Use -r to disable this behavior or -a to force
    it for variables other than PATH and PATHEXT.

    If append is prescribed, but the value doesn't exist, the value will be
    created.

    If there is no value, %prog will delete the <name> environment variable.
    i.e. "PATH="

    To remove a specific value or values from a semicolon-separated
    multi-value variable (such as PATH), use --remove-value.

    e.g. enver --remove-value PATH=C:\\Unwanted\\Dir\\In\\Path

    Remove-value matches case-insensitive and also matches any substring
    so the following would also be sufficient to remove the aforementioned
    undesirable dir.

    enver --remove-value PATH=UNWANTED

    Note that %prog does not affect the current running environment, and can
    only affect subsequently spawned applications.
    """
    from optparse import OptionParser

    parser = OptionParser(usage=trim(enver.__doc__))
    parser.add_option(
        '-U',
        '--user-environment',
        action='store_const',
        const=UserRegisteredEnvironment,
        default=MachineRegisteredEnvironment,
        dest='class_',
        help="Use the current user's environment",
    )
    parser.add_option(
        '-a',
        '--append',
        action='store_true',
        default=False,
        help="Append the value to any existing value (default for PATH and PATHEXT)",
    )
    parser.add_option(
        '-r',
        '--replace',
        action='store_true',
        default=False,
        help="Replace any existing value (used to override default append "
        "for PATH and PATHEXT)",
    )
    parser.add_option(
        '--remove-value',
        action='store_true',
        default=False,
        help="Remove any matching values from a semicolon-separated "
        "multi-value variable",
    )
    parser.add_option(
        '-e',
        '--edit',
        action='store_true',
        default=False,
        help="Edit the value in a local editor",
    )
    options, args = parser.parse_args(*args)

    try:
        param = args.pop()
        if args:
            parser.error("Too many parameters specified")
            raise SystemExit(1)
        if '=' not in param and not options.edit:
            parser.error("Expected <name>= or <name>=<value>")
            raise SystemExit(2)
        name, sep, value = param.partition('=')
        method_name = 'set'
        if options.remove_value:
            method_name = 'remove_values'
        if options.edit:
            method_name = 'edit'
        method = getattr(options.class_, method_name)
        method(name, value, options)
    except IndexError:
        options.class_.show()


if __name__ == '__main__':
    enver()
